package com.danielchen.demo.appreceiver;

import android.support.annotation.NonNull;
import android.util.Log;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;

import okhttp3.Interceptor;
import okhttp3.logging.HttpLoggingInterceptor;

/**
 * 工具集
 *
 * @author danielchen
 * @date 2018/9/30
 */
public class Util {

    private static ThreadLocal<SimpleDateFormat> SDF = new ThreadLocal<SimpleDateFormat>(){

        private static final String TIME_FORMAT = "yyyy-MM-dd HH:mm:ss";

        @Override
        protected SimpleDateFormat initialValue() {
            return new SimpleDateFormat(TIME_FORMAT, Locale.CHINA);
        }
    };

    public static String format(long time) {
        return SDF.get().format(time);
    }

    public static String format(Date time) {
        return SDF.get().format(time);
    }

    public static Date parse(String time) throws ParseException {
        return SDF.get().parse(time);
    }


    private static ThreadLocal<Interceptor> mHttpInterceptor = new ThreadLocal<Interceptor>(){

        @Override
        protected Interceptor initialValue() {
            HttpLoggingInterceptor interceptor = new HttpLoggingInterceptor(new HttpLoggingInterceptor.Logger() {
                @Override
                public void log(@NonNull String message) {
                    Log.e("HTTP","HTTP === " + message);
                }
            });
            interceptor.setLevel(HttpLoggingInterceptor.Level.BODY);
            return interceptor;
        }
    };

    public static Interceptor getHttpInterceptor() {
        return mHttpInterceptor.get();
    }
}
