package com.danielchen.demo.appreceiver;

import android.app.PendingIntent;
import android.content.Context;
import android.content.Intent;
import android.graphics.PixelFormat;
import android.os.Build;
import android.view.Gravity;
import android.view.MotionEvent;
import android.view.View;
import android.view.WindowManager;
import android.widget.Toast;

/**
 * Created by Leinlix on 2018/9/14.
 */

public class ViewManager {
    private FloatingView floatingView;
    private WindowManager windowManager;
    private WindowManager.LayoutParams floatBallParams;

    // 只能初始化一次
    private boolean isInit = false;

    private static final class SingletonHolder{
        private static volatile ViewManager singleton = new ViewManager();
    }

    private ViewManager(){}

    public static ViewManager getInstance (){
        return SingletonHolder.singleton;
    }

    public void showFloatBall(final Context context){

        if (isInit) {
            return;
        }

        floatingView = new FloatingView(context);
        windowManager = (WindowManager)context.getSystemService(Context.WINDOW_SERVICE);
        if(floatBallParams == null){
            floatBallParams = new WindowManager.LayoutParams();
            floatBallParams.width = floatingView.width;
            floatBallParams.height = floatingView.height;
            floatBallParams.gravity = Gravity.TOP| Gravity.START;

            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
                floatBallParams.type = WindowManager.LayoutParams.TYPE_APPLICATION_OVERLAY;
            } else if (Build.VERSION.SDK_INT > Build.VERSION_CODES.N) {
                floatBallParams.type = WindowManager.LayoutParams.TYPE_PHONE;
            } else {
                floatBallParams.type = WindowManager.LayoutParams.TYPE_TOAST;
            }
            floatBallParams.flags = WindowManager.LayoutParams.FLAG_NOT_FOCUSABLE| WindowManager.LayoutParams.FLAG_NOT_TOUCH_MODAL;
            floatBallParams.format = PixelFormat.RGBA_8888;
        }
        windowManager.addView(floatingView,floatBallParams);
        isInit = true;

        floatingView.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View v) {
                // >> danielchen: 这么写试为了开启应用时不需要等5秒
                Intent intent = new Intent(context, MainActivity.class);
                try {
                    PendingIntent pendingIntent = PendingIntent.getActivity(context, 0, intent, 0);
                    pendingIntent.send();
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        });
        floatingView.setOnTouchListener(new View.OnTouchListener() {
            float startX;
            float startY;
            float tempX;
            float tempY;

            @Override
            public boolean onTouch(View v, MotionEvent event) {
                switch (event.getAction()) {
                    case MotionEvent.ACTION_DOWN:
                        startX = event.getRawX();
                        startY = event.getRawY();

                        tempX = event.getRawX();
                        tempY = event.getRawY();
                        break;
                    case MotionEvent.ACTION_MOVE:
                        float dx = event.getRawX() - startX;
                        float dy = event.getRawY() - startY;
                        //计算偏移量，刷新视图
                        floatBallParams.x += dx;
                        floatBallParams.y += dy;
                        windowManager.updateViewLayout(floatingView, floatBallParams);
                        startX = event.getRawX();
                        startY = event.getRawY();
                        break;
                    case MotionEvent.ACTION_UP:
                        //判断松手时View的横坐标是靠近屏幕哪一侧，将View移动到依靠屏幕
                        float endX = event.getRawX();
                        float endY = event.getRawY();
                        if (endX < getScreenWidth() / 2) {
                            endX = 0;
                        } else {
                            endX = getScreenWidth() - floatingView.width;
                        }
                        floatBallParams.x = (int) endX;
                        windowManager.updateViewLayout(floatingView, floatBallParams);
                        //如果初始落点与松手落点的坐标差值超过6个像素，则拦截该点击事件
                        //否则继续传递，将事件交给OnClickListener函数处理
                        if (Math.abs(endX - tempX) > 6 && Math.abs(endY - tempY) > 6) {
                            return true;
                        }
                        break;
                }
                return false;
            }

        });
    }

    private int getScreenWidth() {
        return windowManager.getDefaultDisplay().getWidth();
    }

}
